import os
import sys

from jinjaform import log


args = sys.argv[1:]
cmd = args[0] if args else None
cwd = os.getcwd()
env = os.environ.copy()


def find_project_root():
    current = cwd
    while True:
        path = os.path.join(current, '.jinjaformrc')
        if os.path.exists(path):
            return current
        current = os.path.dirname(current)
        if current == os.path.sep:
            return ''


def find_terraform_bin():
    for path in os.environ['PATH'].split(os.pathsep):
        terraform_path = os.path.join(path, 'terraform')
        if not os.path.exists(terraform_path):
            continue
        if not os.access(terraform_path, os.X_OK):
            continue
        real_name = os.path.basename(os.path.realpath(terraform_path))
        if real_name == 'jinjaform':
            continue
        return terraform_path
    log.bad('terraform: command not found')
    sys.exit(1)


project_root = find_project_root()
jinjaform_root = os.path.join(project_root, '.jinjaform')
terraform_bin = find_terraform_bin()
workspace_dir = os.path.join(cwd, '.jinjaform')
terraform_dir = os.path.join(workspace_dir, '.terraform')

env['JINJAFORM_PROJECT_ROOT'] = project_root
env['JINJAFORM_WORKSPACE'] = workspace_dir



import colorama


def init(_cache=[]):
    if not _cache:
        colorama.init()
        _cache.append(True)


def accept(question, *args, **kwargs):
    if args or kwargs:
        question = question.format(*args, **kwargs)
    question = '[jinjaform] ' + question + ' [yes/no]: '
    answer = ''
    while answer not in ('yes', 'no'):
        try:
            answer = input(question).lower()
        except KeyboardInterrupt:
            answer = 'no'
            print()
    return answer == 'yes'


def bad(message, *args, **kwargs):
    init()
    if args or kwargs:
        message = message.format(*args, **kwargs)
    print(colorama.Fore.RED + '[jinjaform] ' + message + colorama.Style.RESET_ALL)


def ok(message, *args, **kwargs):
    init()
    if args or kwargs:
        message = message.format(*args, **kwargs)
    print(colorama.Fore.CYAN + '[jinjaform] ' + message + colorama.Style.RESET_ALL)



import boto_source_profile_mfa
import boto3
import sys
import threading

from functools import lru_cache

from jinjaform import log
from jinjaform.config import env


aws_provider = {}
s3_backend = {}

lock = threading.Lock()


@lru_cache()
def _get_session(**kwargs):
    if 'profile_name' in kwargs:
        return boto_source_profile_mfa.get_session(**kwargs)
    else:
        return boto3.Session(**kwargs)


def get_default_session():
    session_kwargs = {}
    if aws_provider:
        terraform_boto_session_map = {
            'access_key': 'aws_access_key_id',
            'secret_key': 'aws_secret_access_key',
            'token': 'aws_session_token',
            'profile': 'profile_name',
            'region': 'region_name',
        }
        for terraform_key, boto_key in terraform_boto_session_map.items():
            value = aws_provider.get(terraform_key)
            if value:
                session_kwargs[boto_key] = value
    return get_session(**session_kwargs)


def get_session(**kwargs):
    with lock:
        return _get_session(**kwargs)


def backend_setup():

    region = s3_backend.get('region')
    if not region:
        return

    bucket = s3_backend.get('bucket')
    if bucket:

        log.ok('backend: s3://{} in {}', bucket, region)

        s3_client = get_default_session().client('s3')

        try:
            response = s3_client.get_bucket_versioning(
                Bucket=bucket
            )
        except s3_client.exceptions.NoSuchBucket:
            bucket_exists = False
            bucket_versioning = False
        else:
            bucket_exists = True
            bucket_versioning = response['Status'] == 'Enabled'

        if not bucket_exists:

            if not log.accept('backend: create s3://{} in {}', bucket, region):
                log.bad('backend: bucket not created')
                sys.exit(1)

            log.ok('backend: creating bucket')
            s3_client.create_bucket(
                ACL='private',
                Bucket=bucket,
                CreateBucketConfiguration={
                    'LocationConstraint': region,
                },
            )
            s3_client.get_waiter('bucket_exists').wait(Bucket=bucket)

        if not bucket_versioning:
            log.ok('backend: enabling versioning')
            s3_client.put_bucket_versioning(
                Bucket=bucket,
                VersioningConfiguration={
                    'Status': 'Enabled',
                },
            )

    dynamodb_table = s3_backend.get('dynamodb_table')
    if dynamodb_table:

        log.ok('backend: dynamodb://{} in {}', dynamodb_table, region)

        dynamodb_client = get_default_session().client('dynamodb')

        try:
            dynamodb_client.describe_table(
                TableName=dynamodb_table,
            )
        except dynamodb_client.exceptions.ResourceNotFoundException:

            if not log.accept('backend: create dynamodb://{} in {}', dynamodb_table, region):
                log.bad('backend: table not created')
                sys.exit(1)

            log.ok('creating table')
            dynamodb_client.create_table(
                TableName=dynamodb_table,
                AttributeDefinitions=[
                    {
                        'AttributeName': 'LockID',
                        'AttributeType': 'S',
                    }
                ],
                KeySchema=[
                    {
                        'AttributeName': 'LockID',
                        'KeyType': 'HASH',
                    },
                ],
                BillingMode='PAY_PER_REQUEST',
            )


def credentials_setup():
    """
    Sets up AWS credentials using Terraform AWS provider blocks.
    """

    if not aws_provider:
        return

    session = get_default_session()

    try:
        creds = session.get_credentials().get_frozen_credentials()
    except KeyboardInterrupt:
        print()
        log.bad('aborted')
        sys.exit(1)

    if creds.access_key:
        env['AWS_ACCESS_KEY_ID'] = creds.access_key

    if creds.secret_key:
        env['AWS_SECRET_ACCESS_KEY'] = creds.secret_key

    if creds.token:
        env['AWS_SECURITY_TOKEN'] = creds.token
        env['AWS_SESSION_TOKEN'] = creds.token

    if session.region_name:
        env['AWS_REGION'] = session.region_name
        env['AWS_DEFAULT_REGION'] = session.region_name



import os
import subprocess
import sys

from jinjaform import aws, git, log, rc, terraform, workspace, __version__
from jinjaform.config import args, cmd, cwd, env, project_root, terraform_bin, workspace_dir


commands_bypassed = (
    'fmt',
    'help', '-h', '-help', '--help',
    'version', '-v', '-version', '--version',
)

commands_forbidden = (
    'push',
)


def main():

    if cmd == 'create':
        sys.exit(rc.create())

    if cmd in ('version', '-v', '-version', '--version'):
        log.ok('version: {}'.format(__version__))

    workspace_required = False

    if cmd:

        if cmd in commands_forbidden:
            log.bad('{} is disabled in jinjaform', cmd)
            sys.exit(1)

        if cmd not in commands_bypassed:

            if not project_root:
                log.bad('could not find .jinjaformrc file in current or parent directories')
                log.bad('to start a new jinjaform project in the current directory, run "jinjaform create"')
                sys.exit(1)

            workspace_required = True

    if workspace_required:

        if cwd == project_root:
            log.bad('cannot run from the jinjaform project root directory, aborting')
            sys.exit(1)

        for rc_cmd, rc_arg in rc.read():

            if rc_cmd == 'GIT_CHECK_BRANCH':

                git.check_branch(desired=rc_arg)

            elif rc_cmd == 'GIT_CHECK_CLEAN':

                git.check_clean()

            elif rc_cmd == 'GIT_CHECK_REMOTE':

                git.check_remote()

            elif rc_cmd == 'RUN':

                log.ok('run: {}'.format(rc_arg))
                returncode = subprocess.call(rc_arg, env=env, shell=True)
                if returncode != 0:
                    sys.exit(returncode)

            elif rc_cmd == 'TERRAFORM_RUN':

                log.ok('run: terraform')
                os.chdir(workspace_dir)
                returncode = terraform.execute(terraform_bin, args, env)
                if returncode != 0:
                    sys.exit(returncode)

            elif rc_cmd == 'WORKSPACE_CREATE':

                workspace.clean()
                workspace.create()

                aws.credentials_setup()

                if cmd == 'init':
                    aws.backend_setup()

            else:

                log.bad('configuration: {} is not a valid command', rc_cmd)
                sys.exit(1)

    else:

        log.ok('run: terraform')
        returncode = terraform.execute(terraform_bin, args, env)
        if returncode != 0:
            sys.exit(returncode)


if __name__ == '__main__':
    main()



import hcl
import importlib
import os
import pkgutil
import re
import shutil
import sys

from collections import defaultdict
from contextlib import suppress

from functools import partial

from getpass import getpass

from itertools import chain

from jinja2 import Environment, StrictUndefined
from jinja2.exceptions import UndefinedError

from jinjaform import aws, log
from jinjaform.config import cwd, env, jinjaform_root, project_root, terraform_dir, workspace_dir

from queue import Queue

from threading import current_thread, Event, Lock, Thread


class Prompter(object):
    """
    Helper class that allows prompts from background threads
    to occur in the main thread. This avoids issues with
    pressing Ctrl-C and having threads fail in the background.

    """

    def __init__(self):
        self._queue = Queue()
        self._prompts = {}
        self._results = {}
        self._running = True

    def prompt(self, prompt):
        event = Event()
        self._prompts[event] = prompt
        self._queue.put(event)
        event.wait()
        result = self._results.get(event, '')
        if result is KeyboardInterrupt:
            print()
            raise KeyboardInterrupt
        return result

    def start(self):
        while True:
            event = self._queue.get()
            if not event:
                return
            if self._running:
                prompt = self._prompts[event]
                try:
                    result = getpass(prompt)
                except KeyboardInterrupt:
                    result = KeyboardInterrupt
                self._results[event] = result
            event.set()

    def stop(self):
        self._running = False
        self._queue.put(None)


class VarStore(object):

    def __init__(self, threads):

        self._running = True
        self._lock = Lock()
        self._threads = threads
        self._threads_waiting = defaultdict(dict)
        self._variables_waiting = defaultdict(list)

        self._defined = set()
        self._defaults = dict()
        self._values = dict()

        self._unresolved = defaultdict(set)

    def __getitem__(self, key):
        """
        Returns a variable value. If the variable has not been defined
        then this will block and wait for it.

        """

        # Wait if the variable has not been defined yet.
        self._wait_for_variable(key)

        # Return the value if the variable has been defined
        # with a specific value or a default value.
        if key in self._defined:
            if key in self._values:
                return self._values[key]
            if key in self._defaults:
                return self._defaults[key]

        # The variable was not defined, or it was defined without a value,
        # or it was defined but there was a deadlock when resolving it.
        self._unresolved[current_thread()].add(key)
        raise KeyError(key)

    def _cancel(self):
        """
        This function will unblock all threads waiting on a variable,
        allowing them to continue executing without their variable.
        They will then have an error because the variable isn't defined.

        """

        self._running = False
        for event in chain.from_iterable(self._variables_waiting.values()):
            event.set()

    def _check_deadlock(self):
        """
        Checks if there are any free threads. If all threads are blocked
        waiting for variables to be defined, then no more variables will
        be defined and the threads will never finish rendering. In that
        case, cancel rendering and allow errors to be raised.

        """

        for thread in self._get_free_threads():
            break
        else:
            self._cancel()

    def _define_variable(self, name, default):
        """
        Adds a variable definition and unblocks any waiting threads.

        """

        with self._lock:
            self._defined.add(name)
            if default is not None:
                self._defaults[name] = default
            for event in self._variables_waiting.pop(name, []):
                event.set()

    def _get_free_threads(self):
        """
        Returns all threads which are not currently waiting for a variable
        to be defined. These are important when determining if there is
        a deadlock; if all threads are blocked waiting for variables to
        be defined, then no more variables will be define and the threads
        will never unblock by themselves.

        """

        # Return nothing if the rendering has been cancelled.
        if not self._running:
            return

        # Check all running threads.
        for thread in self._threads:
            for name in self._threads_waiting[thread].values():
                if name not in self._defined:
                    # This thread is waiting on a variable which hasn't been
                    # defined yet, so it is not free.
                    break
            else:
                # This thread is no waiting on any variables that haven't
                # been defined yet, so it is free.
                yield thread

    def _get_unresolved_variables(self):
        return sorted(self._unresolved.get(current_thread(), set()))

    def _set_variable_value(self, name, value):
        self._values[name] = value

    def _thread_done(self):
        """
        Removes the current thread and checks for a deadlock.

        """

        with self._lock:
            self._threads.remove(current_thread())
            self._check_deadlock()

    def _wait_for_variable(self, name):
        """
        Waits for a variable to be defined in another template being rendered
        by another thread. Returns immediately if it is already defined.

        """

        with self._lock:

            if name in self._defined:
                return

            # Check if there are any other threads that aren't waiting on
            # a variable. These other threads may end up defining this
            # variable, or they may not. If there are no free threads
            # then there is no way for the variable to be defined, so
            # cancel the rendering and allow an error to be raised.
            for thread in self._get_free_threads():
                if thread != current_thread():
                    break
            else:
                self._cancel()
                return

            # Create an event that will be used to block execution
            # in this thread until another thread unblocks it.
            event = Event()

            # Add a reference to the event by the variable name.
            # This allows another thread that defines the variable
            # to find this thread and unblock it.
            self._variables_waiting[name].append(event)

            # Add a reference to the event by the thread.
            # This allows other threads to check if this
            # thread is blocked or not.
            self._threads_waiting[current_thread()][event] = name

        # Wait for the variable to be defined,
        # or for rendering to be cancelled.
        event.wait()

        with self._lock:

            # Remove the reference so this thread can be seen as free.
            self._threads_waiting[current_thread()].pop(event)

            # Check if the rendering is in a deadlock.
            self._check_deadlock()


class MultiTemplateRenderer(object):

    def __init__(self):
        self._threads = set()
        self._var_store = VarStore(self._threads)
        self._errors = []
        self._prompter = Prompter()
        self._rendered = {}
        self._jinja_context = {}
        self._jinja_environment = self._create_jinja_environment()

    def _create_jinja_environment(self):

        # Create a Jina2 Environment.
        env = Environment(
            undefined=StrictUndefined,
            keep_trailing_newline=True,
            extensions=[
                'jinja2.ext.do',
                'jinja2.ext.loopcontrols',
            ],
        )

        # Create a context for the template to use.
        # Include environment variables and the var store which
        # exposes the `var.some_name` Terraform variables.
        # Custom context values will be added below too.
        self._jinja_context.update({
            'aws': {
                'session': partial(aws.get_session, mfa_prompter=self._prompter.prompt),
            },
        })
        self._jinja_context.update(os.environ)
        self._jinja_context['var'] = self._var_store

        # Load Jinja2 extensions.
        jinja_path = os.path.join(project_root, '.jinja')
        if os.path.exists(jinja_path):
            sys.path.insert(0, jinja_path)

            context_path = os.path.join(jinja_path, 'context')
            for module_finder, name, ispkg in pkgutil.iter_modules(path=[context_path]):
                module = importlib.import_module('context.' + name)
                for name in getattr(module, '__all__', []):
                    self._jinja_context[name] = getattr(module, name)

            filters_path = os.path.join(jinja_path, 'filters')
            for module_finder, name, ispkg in pkgutil.iter_modules(path=[filters_path]):
                module = importlib.import_module('filters.' + name)
                for name in getattr(module, '__all__', []):
                    env.filters[name] = getattr(module, name)

            tests_path = os.path.join(jinja_path, 'tests')
            for module_finder, name, ispkg in pkgutil.iter_modules(path=[tests_path]):
                module = importlib.import_module('tests.' + name)
                for name in getattr(module, '__all__', []):
                    env.tests[name] = getattr(module, name)

        return env

    def _render(self, source):
        try:

            # Render the template.
            with open(source) as open_file:
                template = self._jinja_environment.from_string(open_file.read())
            try:
                rendered = template.render(**self._jinja_context)
            except UndefinedError as error:
                names = self._var_store._get_unresolved_variables()
                for name in names:
                    error = "'var.{}' cannot be resolved".format(name)
                    self._errors.append('{} in {}'.format(error, source))
                if not names:
                    self._errors.append('{} in {}'.format(error, source))
                return
            except KeyboardInterrupt:
                self._errors.append('interrupted')
                return

            # Save rendered templates.
            self._rendered[source] = rendered

            # Parse config.
            try:
                parsed = hcl.loads(rendered)
            except ValueError:
                parsed = {}
                in_comment = False
                for line in rendered:
                    line = line.lstrip()
                    if not line:
                        continue
                    if line.startswith('/*'):
                        in_comment = True
                    elif in_comment:
                        if line.startswith('*/'):
                            in_comment = False
                        elif line.startswith('#'):
                            continue
                        else:
                            log.bad('error parsing {}', source)
                            break

            # Process variables.
            variables = parsed.get('variable', {})
            for name, data in variables.items():
                default = data.get('default', None)
                self._var_store._define_variable(name, default)

            # Process the default AWS provider.
            provider = parsed.get('provider')
            if provider and provider.get('aws'):
                # pyhcl is broken and merges blocks
                # so parse them with a custom function.
                for aws_provider in _parse_aws_providers(rendered):
                    if not aws_provider.get('alias'):
                        aws.aws_provider.update(aws_provider)

            # Process the S3 backend.
            terraform = parsed.get('terraform')
            if terraform:
                backend = terraform.get('backend')
                if backend:
                    s3 = backend.get('s3')
                    if s3:
                        aws.s3_backend.update(s3)

        except Exception as error:
            self._errors.append('{} in {}'.format(error, source))
        finally:
            self._var_store._thread_done()
            if not self._threads:
                self._prompter.stop()

    def add_template(self, source):
        self._threads.add(Thread(
            target=self._render,
            kwargs={'source': source},
        ))

    def set_variable_value(self, name, value):
        self._var_store._set_variable_value(name, value)

    def start(self):
        threads = self._threads.copy()
        for thread in threads:
            thread.start()
        self._prompter.start()
        for thread in threads:
            thread.join()
        for error in self._errors:
            log.bad(error)
        success = not bool(self._errors)
        return (success, self._rendered)


def _parse_aws_providers(rendered):
    blocks = []
    buffer = []
    for line in rendered.splitlines():
        stripped_line = line.strip()
        if blocks:
            buffer.append(line)
            if stripped_line == '}':
                blocks.pop()
                if not blocks:
                    parsed = hcl.loads('\n'.join(buffer))
                    yield parsed['provider']['aws']
                    buffer.clear()
            else:
                match = re.match(r'^([a-z]+)\s+\{$', stripped_line)
                if match:
                    blocks.append(match.group(1))
        else:
            match = re.match(r'^(provider\s+"aws")\s+\{$', stripped_line)
            if match:
                blocks.append(match.group(1))
                buffer.append(line)


def _populate():

    # Create a template renderer that can handle multiple files
    # with variable references between files.
    template_renderer = MultiTemplateRenderer()

    # Discover files to create in the workspace. Files in multiple
    # levels of the project directory tree with the same name will
    # be combined into a single file in the workspace.

    tfvars_files = defaultdict(set)
    tf_files = defaultdict(set)
    other_files = defaultdict(set)

    current = cwd
    while (current + '/').startswith(project_root + '/'):
        for name in sorted(os.listdir(current)):
            if name.startswith('.'):
                continue
            path = os.path.join(current, name)
            if os.path.isdir(path):
                continue
            name = name.lower()
            if name.endswith('.tfvars'):
                tfvars_files[name].add(path)
            elif name.endswith('.tf'):
                tf_files[name].add(path)
            else:
                other_files[name].add(path)
        current = os.path.dirname(current)

    # Process .tfvars files first, and read their variable values,
    # because they are required when rendering .tf files.

    for name in sorted(tfvars_files):

        source_paths = sorted(tfvars_files[name])
        target_path = os.path.join(workspace_dir, name)

        if len(source_paths) == 1:
            log.ok('copy: {}', name)
        else:
            log.ok('combine: {}', name)

        with open(target_path, 'w') as output_file:

            for source_path in source_paths:

                with open(source_path) as source_file:
                    source_file_contents = source_file.read()

                relative_source_path = os.path.relpath(source_path, project_root)
                output_file.write('# jinjaform: {}'.format(relative_source_path))
                output_file.write('\n\n')
                output_file.write(source_file_contents)
                output_file.write('\n')

                if name == 'terraform.tfvars':
                    for key, value in hcl.loads(source_file_contents).items():
                        template_renderer.set_variable_value(key, value)

    # Process .tf files as templates.

    for name in sorted(tf_files):

        source_paths = sorted(tf_files[name])

        log.ok('render: {}', name)

        for source_path in source_paths:
            template_renderer.add_template(source_path)

    success, rendered = template_renderer.start()
    if not success:
        sys.exit(1)

    for name in sorted(tf_files):

        source_paths = sorted(tf_files[name])
        target_path = os.path.join(workspace_dir, name)

        with open(target_path, 'w') as output_file:

            for source_path in source_paths:

                relative_source_path = os.path.relpath(source_path, project_root)
                output_file.write('# jinjaform: {}'.format(relative_source_path))
                output_file.write('\n\n')
                output_file.write(rendered[source_path])
                output_file.write('\n')

    # Process remaining files. Do not add source comments because
    # the file format is unknown (e.g. json files would break with #).
    for name in sorted(other_files):

        source_paths = sorted(other_files[name])
        target_path = os.path.join(workspace_dir, name)

        if len(source_paths) == 1:
            log.ok('copy: {}', name)
        else:
            log.ok('combine: {}', name)

        with open(target_path, 'w') as output_file:
            for source_path in source_paths:
                with open(source_path) as source_file:
                    output_file.write(source_file.read())


def _remove(path):
    with suppress(FileNotFoundError):
        if os.path.islink(path):
            os.remove(path)
        if os.path.isdir(path):
            shutil.rmtree(path)
        else:
            os.remove(path)


def clean():
    if os.path.exists(workspace_dir):
        for name in os.listdir(workspace_dir):
            if name != '.terraform':
                _remove(os.path.join(workspace_dir, name))


def create():
    # Ensure the .jinjaform/.terraform directory exists.
    os.makedirs(terraform_dir, exist_ok=True)

    # Create a shared modules directory for the entire project.
    module_cache_dir = os.path.join(jinjaform_root, 'modules')
    os.makedirs(module_cache_dir, exist_ok=True)
    module_link = os.path.join(terraform_dir, 'modules')
    _remove(module_link)
    os.symlink(module_cache_dir, module_link)

    # Create a shared plugin cache directory for the entire project.
    plugin_cache_dir = os.path.join(jinjaform_root, 'plugins')
    os.makedirs(plugin_cache_dir, exist_ok=True)
    env['TF_PLUGIN_CACHE_DIR'] = plugin_cache_dir

    # Populate workspace with Terraform configuration files.
    _populate()



import os

from jinjaform import log
from jinjaform.config import cwd, project_root


default = '''
# Check if the git checkout is clean.
# Only runs when using the "terraform apply" command.
GIT_CHECK_CLEAN

# Check if the master branch is checked out.
# Only runs when using the "terraform apply" command.
GIT_CHECK_BRANCH master

# Check if the local branch is up to date.
# Only runs when using the "terraform apply" command.
GIT_CHECK_REMOTE

# Create the Jinjaform workspace.
# Runs for all terraform commands except: help, fmt, version
WORKSPACE_CREATE

# Run terraform.
TERRAFORM_RUN
'''.lstrip()


def create():
    path = os.path.join(cwd, '.jinjaformrc')
    try:
        with open(path, 'x') as open_file:
            open_file.write(default)
    except FileExistsError:
        log.bad('.jinjaformrc file already exists')
        return 1
    else:
        log.ok('created {}', path)
        log.ok('your project root directory is {}', cwd)
        return 0


def read():

    rc_path = os.path.join(project_root, '.jinjaformrc')
    commands = []
    with open(rc_path) as open_file:
        for line in open_file:
            line = line.strip()
            if line and not line.startswith('#'):
                commands.append(line)

    if commands.count('WORKSPACE_CREATE') != 1:
        log.bad('configuration: WORKSPACE_CREATE must be defined once')

    if commands.count('TERRAFORM_RUN') != 1:
        log.bad('configuration: TERRAFORM_RUN must be defined once')

    if commands.index('TERRAFORM_RUN') < commands.index('WORKSPACE_CREATE'):
        log.bad('configuration: WORKSPACE_CREATE must be defined before TERRAFORM_RUN')

    for command in commands:
        parts = command.split(None, 1)
        if len(parts) == 2:
            yield parts
        else:
            yield command, None



import errno
import os


def execute(terraform_bin, args, env):
    child_pid = os.fork()
    if child_pid == 0:
        os.execvpe(terraform_bin, [terraform_bin] + args, env)
    else:
        while True:
            try:
                _, exit_status = os.waitpid(child_pid, 0)
            except KeyboardInterrupt:
                pass
            except OSError as error:
                if error.errno == errno.ECHILD:
                    # No child processes.
                    # It has exited already.
                    break
                elif error.errno == errno.EINTR:
                    # Interrupted system call.
                    # This happens when resizing the terminal.
                    pass
                else:
                    # An actual error occurred.
                    raise
            else:
                return exit_status
